/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using System.IO;

namespace BoxEdit3D
{
    /// <summary>
    /// This class encapsulates on box in the scene.
    /// </summary>
    class Box
    {
        // internal variables to hold the position, scale, rotation and 
        // the path to the texture
        private Vector3 m_translation;
        private Vector3 m_scale;
        private Quaternion m_rotation;
        private string m_textureFilename;

        // if true this box renders a translucent outer cube around itself
        private bool m_isSelected = false;

        // if true, the matrices have to be updated
        private bool m_updateMatrix = true;

        // matrix that combines the translation, scale and rotation of the box
        private Matrix m_modelMatrix;
        private Matrix m_modelMatrixWithoutRotation;

        // only use one model for all the cubes
        private static Model m_cube;
        private static Model m_selectedCube;

        // give read only access to the model so that the picking 
        // can do per triangle tests
        public static Model cubeModel
        {
            get
            {
                return m_cube;
            }
        }

        // effects used to render the box
        private static Effect m_cubeEffect;
        private static Effect m_selectedCubeEffect;

        // texture manager that holds all the textures, so that
        // no texture has to be stored twice
        private static TextureManager m_textureManager;

        // needed to set some renderstates during rendering
        private GraphicsDeviceManager m_graphics;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="graphics">GraphicsDeviceManager to obtain the GraphicsDevice during rendering.</param>
        public Box(GraphicsDeviceManager graphics)
        {
            m_graphics = graphics;
            m_translation = Vector3.Zero;
            m_scale = Vector3.One;
            m_rotation = Quaternion.Identity;
            m_textureFilename = "";
            m_isSelected = false;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Constructor.
        /// Create Box from values save in the XMLBox class provided.
        /// </summary>
        /// <param name="graphics">GraphicsDeviceManager to obtain the GraphicsDevice during rendering.</param>
        /// <param name="xmlBox">The XMLBox which values should be used to initialize this class.</param>
        public Box(GraphicsDeviceManager graphics, XMLBox xmlBox)
        {
            m_graphics = graphics;
            m_translation = Vector3.Zero;
            m_scale = Vector3.One;
            m_rotation = Quaternion.Identity;
            m_textureFilename = "";
            m_isSelected = false;
            m_updateMatrix = true;

            // use set methods to check for correct values
            setTranslation(xmlBox.translation);
            setScale(xmlBox.scale);
            setRotation(xmlBox.rotation);
            setTextureFilename(xmlBox.texturePath);
        }

        /// <summary>
        /// Returns an XMLBox object filled with the information of this box.
        /// </summary>
        /// <returns>XMLBox instance filled with internal data.</returns>
        internal XMLBox getXMLBox()
        {
            XMLBox xmlBox = new XMLBox();
            xmlBox.translation = m_translation;
            xmlBox.scale = m_scale;
            xmlBox.rotation = m_rotation;
            xmlBox.texturePath = m_textureFilename;
            return xmlBox;
        }

        /// <summary>
        /// Recalculates the model matrices if necessary.
        /// </summary>
        private void updateMatrix() {
            if (m_updateMatrix) {
                m_modelMatrixWithoutRotation = Matrix.CreateScale(m_scale) * Matrix.CreateTranslation(m_translation);
                m_modelMatrix = Matrix.CreateScale(m_scale) * Matrix.CreateFromQuaternion(m_rotation) * Matrix.CreateTranslation(m_translation); 
                m_updateMatrix = false;
            }
        }
        
        /// <summary>
        /// Returns the tanslation of the box.
        /// </summary>
        /// <returns>The absolute translation of the box in World Coordinates.</returns>
        public Vector3 getTranslation()
        {
            return m_translation;
        }

        /// <summary>
        /// Set the translation of the box.
        /// </summary>
        /// <param name="translation">The new absolute translation of the box in World Coordinates.</param>
        public void setTranslation(Vector3 translation)
        {
            m_translation = translation;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Translates the box by the given Vector3.
        /// </summary>
        /// <param name="translation">The Vector3 of the relative translation in World Coordinates.</param>
        public void addTranslation(Vector3 translation)
        {
            m_translation += translation;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Returns the Scale of the box.
        /// </summary>
        /// <returns>Vector with the scale in X, Y and Z-direction.</returns>
        public Vector3 getScale()
        {
            return m_scale;
        }

        /// <summary>
        /// Sets the scale of the box. All scales have to greater zero.
        /// </summary>
        /// <param name="scale">Vector with the scale in X, Y and Z-direction.</param>
        public void setScale(Vector3 scale)
        {
            if ((scale.X <= 0) || (scale.Y <= 0) || (scale.Z <= 0)) return;
            m_scale = scale;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Scales the box relative to the current scale. All scales have to greater zero.
        /// </summary>
        /// <param name="scale">Vector with the relative scale in X, Y and Z-direction.</param>
        public void scaleBy(Vector3 scale)
        {
            if ((scale.X <= 0) || (scale.Y <= 0) || (scale.Z <= 0)) return;
            m_scale *= scale;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Returns the rotation of the box.
        /// </summary>
        /// <returns>A Quaternion that represents the rotation of the box.</returns>
        public Quaternion getRotation()
        {
            return m_rotation;
        }

        /// <summary>
        /// Sets the rotation of the box.
        /// </summary>
        /// <param name="rotation">A Quaternion that represents the rotation of the box.</param>
        public void setRotation(Quaternion rotation)
        {
            m_rotation = rotation;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Adds the supplied rotation to the current rotation of the box.
        /// </summary>
        /// <param name="rotation">A Quaternion that represents the relative rotation of the box.</param>
        public void addRotation(Quaternion rotation)
        {
            m_rotation = Quaternion.Concatenate(m_rotation, rotation);
            m_updateMatrix = true;
        }

        /// <summary>
        /// Returns if the current box is selected and will display a translucent hull around itself.
        /// </summary>
        /// <returns>True if this box is selected by the user.</returns>
        public bool getSelectedStatus()
        {
            return m_isSelected;
        }

        /// <summary>
        /// Sets the status if this box is selected by the user.
        /// </summary>
        /// <param name="selected">Set this value to true if the box is selected by the user.</param>
        public void setSelectedStatus(bool selected)
        {
            m_isSelected = selected;
        }

        /// <summary>
        /// Returns the path and filename of the texture that is used when drawing this box.
        /// If this value is an empty string, the box will be drawn without a texture.
        /// </summary>
        /// <returns>The filename of the texture.</returns>
        public string getTextureFilename()
        {
            return m_textureFilename;
        }

        /// <summary>
        /// Sets the path and filename of the texture that is used when drawing this box.
        /// If this value is an empty string, the box will be drawn without a texture.
        /// </summary>
        /// <param name="filename">The filename of the new texture.</param>
        /// <returns>True if the new filename was excepted.</returns>
        public bool setTextureFilename(string filename)
        {
            if ((filename == "") || (File.Exists(filename))) {
                m_textureFilename = filename;
                return true;
            }
            else return false;
        }

        /// <summary>
        /// This method loads the static Resources.
        /// </summary>
        /// <param name="graphics">The GraphicsDeviceManager needed for the texture manager.</param>
        /// <param name="content">The ContentManager giving access to the resources.</param>
        public static void loadResources(GraphicsDeviceManager graphics, ContentManager content)
        {
            // hold two instances of the cube model to apply two different shader
            m_cube = content.Load<Model>("Models\\cube");
            m_selectedCube = content.Load<Model>("Models\\selectedCube");
            m_cubeEffect = content.Load<Effect>("Shader\\CubeShader");
            m_selectedCubeEffect = content.Load<Effect>("Shader\\SelectedCubeShader");

            // set the effect file for the meshes
            foreach (ModelMesh modmesh in m_cube.Meshes)
            {
                foreach (ModelMeshPart modmeshpart in modmesh.MeshParts)
                {
                    modmeshpart.Effect = m_cubeEffect;
                }
            }
            foreach (ModelMesh modmesh in m_selectedCube.Meshes)
            {
                foreach (ModelMeshPart modmeshpart in modmesh.MeshParts)
                {
                    modmeshpart.Effect = m_selectedCubeEffect;
                }
            }

            // initialize the texture manager
            m_textureManager = new TextureManager(graphics);
        }

        /// <summary>
        /// Unload all resources that have been allocated by the user.
        /// </summary>
        public static void unloadResources()
        {
            m_textureManager.freeResources();
        }

        /// <summary>
        /// Render the cube with the given viewProjection Matrix.
        /// </summary>
        /// <param name="viewProj">The ViewProjection Matrix that should be used during rendering.</param>
        public void render(Matrix viewProj)
        {
            // check if matrix is up to date
            updateMatrix();
            Matrix worldViewProj = m_modelMatrix * viewProj;

            // get transforms for the meshes
            Matrix[] transforms = new Matrix[m_cube.Bones.Count];
            m_cube.CopyAbsoluteBoneTransformsTo(transforms);

            // render each mesh
            foreach (ModelMesh mesh in m_cube.Meshes)
            {
                foreach (Effect effect in mesh.Effects)
                {
                    // use shader to determine if a texture is displayed 
                    effect.Parameters["g_worldViewProj"].SetValue(transforms[mesh.ParentBone.Index] * worldViewProj);
                    effect.Parameters["g_useTexture"].SetValue((m_textureFilename != ""));
                    effect.Parameters["g_texture"].SetValue(m_textureManager.getTexture(m_textureFilename));
                }
                mesh.Draw();
            }

            // draw translucent hull around cube, if selected
            if (m_isSelected) {

                // get transforms for the meshes
                transforms = new Matrix[m_selectedCube.Bones.Count];
                m_selectedCube.CopyAbsoluteBoneTransformsTo(transforms);

                // make the hull slightly bigger than the normal cube
                worldViewProj = Matrix.CreateScale(new Vector3(1.05f, 1.05f, 1.05f)) * m_modelMatrix * viewProj;

                // change render states and save old render states
                CullMode oldMode = m_graphics.GraphicsDevice.RenderState.CullMode;
                bool alpha = m_graphics.GraphicsDevice.RenderState.AlphaBlendEnable;
                m_graphics.GraphicsDevice.RenderState.AlphaBlendEnable = true;
                m_graphics.GraphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha;
                m_graphics.GraphicsDevice.RenderState.DestinationBlend = Blend.InverseSourceAlpha;

                // first render back faces for correct translucent faces
                m_graphics.GraphicsDevice.RenderState.CullMode = CullMode.CullCounterClockwiseFace;
                foreach (ModelMesh mesh in m_selectedCube.Meshes)
                {
                    foreach (Effect effect in mesh.Effects)
                    {
                        effect.Parameters["g_worldViewProj"].SetValue(transforms[mesh.ParentBone.Index] * worldViewProj);
                    }
                    mesh.Draw();
                }

                // now render front faces for correct translucent faces
                m_graphics.GraphicsDevice.RenderState.CullMode = CullMode.CullClockwiseFace;
                foreach (ModelMesh mesh in m_selectedCube.Meshes)
                {
                    foreach (Effect effect in mesh.Effects)
                    {
                        effect.Parameters["g_worldViewProj"].SetValue(transforms[mesh.ParentBone.Index] * worldViewProj);
                    }
                    mesh.Draw();
                }

                // change renderstates back to previous values
                m_graphics.GraphicsDevice.RenderState.AlphaBlendEnable = alpha;
                m_graphics.GraphicsDevice.RenderState.CullMode = oldMode;
            }
        }

        /// <summary>
        /// Returns the transformation matrix for this box.
        /// </summary>
        /// <returns>The Matrix used to change the original model.</returns>
        public Matrix getTransformMatrix()
        {
            updateMatrix();
            return m_modelMatrix;
        }

        /// <summary>
        /// Returns the transformation matrix without the rotation applied.
        /// This matrix will be used to always draw the translation gizmo axis aligned.
        /// </summary>
        /// <returns>The Matrix used to change the original model without the rotation.</returns>
        public Matrix getTransformMatrixWithoutRotation()
        {
            updateMatrix();
            return m_modelMatrixWithoutRotation;
        }
    }
}
