/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Graphics;

namespace BoxEdit3D
{
    /// <summary>
    /// This class encapsulates the camera and its movement around the scene.
    /// </summary>
    class Camera
    {
        // internal variables that describe important parameters of the camera
        private Vector3 m_position;
        private Vector3 m_direction;
        private Vector3 m_up;

        // rotation of the camera
        private float m_yaw = 0.0f;
        private float m_pitch = 0.0f;
        private float m_roll = 0.0f;

        // needed to determine the aspect ratio of the graphics device
        private GraphicsDeviceManager m_graphics;

        // variables so that the matrices don't have to be calculated every
        // time they are accessed
        private Matrix m_viewMatrix;
        private Matrix m_projMatrix;
        private Matrix m_viewProjMatrix;

        // hold the mouse position in the previous frame
        private int m_oldMousePositionX;
        private int m_oldMousePositionY;

        // true if the matrices variables need to be updated
        private bool m_updateMatrix;

        // Velocity of the Camera in units per frame.
        private const float m_velocity = 0.1f;

        // Factor to multiply mouse movement in pixel to an angle.
        private const float m_rotationFactor = MathHelper.Pi / 1440.0f;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="graphics">The GraphicsDeviceManager is needed to determine the aspect ratio of the graphics device.</param>
        public Camera(GraphicsDeviceManager graphics)
        {
            m_graphics = graphics;
            reset();
        }

        /// <summary>
        /// Reset the camera position and rotation to its original values.
        /// </summary>
        public void reset()
        {
            m_position = new Vector3(0.0f, 1.5f, 10.0f);
            m_direction = new Vector3(0.0f, 0.0f, -1.0f);
            m_up = new Vector3(0.0f, 1.0f, 0.0f);

            // no rotation
            m_yaw = 0.0f;
            m_pitch = 0.0f;
            m_roll = 0.0f;

            // no mouse position yet
            m_oldMousePositionX = -1;
            m_oldMousePositionY = -1;

            // matrices need to be recalculated
            m_updateMatrix = true;
        }

        /// <summary>
        /// Move the Camera forward by a fixed value.
        /// </summary>
        public void moveForward()
        {
            // rotation of the camera is only saved in the rotation quaternion
            Vector3 vDir = Vector3.Transform(m_direction, Matrix.CreateFromYawPitchRoll(m_yaw, m_pitch, m_roll));
            m_position += vDir * m_velocity;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Move the Camera backwards by a fixed value.
        /// </summary>
        public void moveBackward()
        {
            // rotation of the camera is only saved in the rotation quaternion
            Vector3 vDir = Vector3.Transform(m_direction, Matrix.CreateFromYawPitchRoll(m_yaw, m_pitch, m_roll));
            m_position += -vDir * m_velocity;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Strafe the camera to the left by a fixed value.
        /// </summary>
        public void strafeLeft()
        {
            // rotation of the camera is only saved in the rotation quaternion
            // create cross product of the rotated direction and up vector
            // to recieve a vector pointing left
            Vector3 vDir = Vector3.Transform(m_direction, Matrix.CreateFromYawPitchRoll(m_yaw, m_pitch, m_roll));
            Vector3 vUp = Vector3.Transform(m_up, Matrix.CreateFromYawPitchRoll(m_yaw, m_pitch, m_roll));
            Vector3 left = Vector3.Cross(vUp, vDir);
            m_position += left * m_velocity;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Strafe the camera to the right by a fixed value.
        /// </summary>
        public void strafeRight()
        {
            // rotation of the camera is only saved in the rotation quaternion
            // create cross product of the rotated direction and up vector
            // to recieve a vector pointing right
            Vector3 vDir = Vector3.Transform(m_direction, Matrix.CreateFromYawPitchRoll(m_yaw, m_pitch, m_roll));
            Vector3 vUp = Vector3.Transform(m_up, Matrix.CreateFromYawPitchRoll(m_yaw, m_pitch, m_roll));
            Vector3 right = -Vector3.Cross(vUp, vDir);
            m_position += right * m_velocity;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Move the camera upwards by a fixed value.
        /// </summary>
        public void moveUp()
        {
            Vector3 vUp = Vector3.Transform(m_up, Matrix.CreateFromYawPitchRoll(m_yaw, m_pitch, m_roll));
            m_position += vUp * m_velocity;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Move the camera downwards by a fixed value.
        /// </summary>
        public void moveDown()
        {
            Vector3 vUp = Vector3.Transform(m_up, Matrix.CreateFromYawPitchRoll(m_yaw, m_pitch, m_roll));
            m_position += -vUp * m_velocity;
            m_updateMatrix = true;
        }

        /// <summary>
        /// Reacts on the mouse movement and rotates the camera if the
        /// right mouse button was pressed.
        /// </summary>
        public void handleMouseInput()
        {
            MouseState mouseState = Mouse.GetState(); 
            // if this is the first time this method gets called
            // save the current mouse position
            if ((m_oldMousePositionX == -1) && (m_oldMousePositionY == -1))
            {
                m_oldMousePositionX = mouseState.X;
                m_oldMousePositionY = mouseState.Y;
            }
            else
            {
                if (mouseState.RightButton == ButtonState.Pressed)
                {
                    // rotate the camera
                    // create rotation matrix
                    int dx = m_oldMousePositionX - mouseState.X;
                    int dy = m_oldMousePositionY - mouseState.Y;

                    // use yaw and pitch to create a quaternion
                    // with the new rotation
                    float yaw;
                    float pitch;
                    if (dx != 0) yaw = (float)dx * m_rotationFactor;
                    else yaw = 0.0f;
                    if (dy != 0) pitch = (float)dy * m_rotationFactor;
                    else pitch = 0;

                    // add new yaw and pitch
                    m_yaw += yaw;
                    m_pitch += pitch;

                    // check for bounds on pitch
                    if (m_pitch > (MathHelper.Pi / 2.0f)) m_pitch = (MathHelper.Pi / 2.0f);
                    if (m_pitch < -(MathHelper.Pi / 2.0f)) m_pitch = -(MathHelper.Pi / 2.0f);
                    m_updateMatrix = true;

                    // save current mouse position
                    m_oldMousePositionX = mouseState.X;
                    m_oldMousePositionY = mouseState.Y;
                }
                else
                {
                    // update old mouse position
                    m_oldMousePositionX = mouseState.X;
                    m_oldMousePositionY = mouseState.Y;
                }
            }
        }

        /// <summary>
        /// Update all matrices if they need to be recalculated.
        /// </summary>
        private void updateMatrix()
        {
            if (m_updateMatrix)
            {
                // calculate aspect ratio of the graphics device
                float aspectRatio = (float)m_graphics.GraphicsDevice.Viewport.Width / (float)m_graphics.GraphicsDevice.Viewport.Height;

                // m_direction and m_up will not be changed but rather be
                // rotated for every update
                Matrix rotation = Matrix.CreateFromYawPitchRoll(m_yaw, m_pitch, m_roll);
                Vector3 direction = Vector3.Transform(m_direction, rotation);
                Vector3 up = Vector3.Transform(m_up, rotation);
                
                // creat projection and view matrix
                m_projMatrix = Matrix.CreatePerspectiveFieldOfView(MathHelper.ToRadians(45.0f), aspectRatio, 0.01f, 1000.0f);
                m_viewMatrix = Matrix.CreateLookAt(m_position, m_position + direction, up);
                m_viewProjMatrix = m_viewMatrix * m_projMatrix;
                m_updateMatrix = false;
            }
        }

        /// <summary>
        /// Returns the View and Projection Matrix combined. Used during the rendering of objects.
        /// </summary>
        /// <returns>The combined view and projection matrix.</returns>
        public Matrix getViewProjectionMatrix()
        {
            updateMatrix();
            return m_viewProjMatrix;
        }

        /// <summary>
        /// Returns the view matrix of the camera.
        /// </summary>
        /// <returns>The view matrix.</returns>
        public Matrix getViewMatrix()
        {
            updateMatrix();
            return m_viewMatrix;
        }

        /// <summary>
        /// Returns the projection matrix of the camera.
        /// </summary>
        /// <returns>The projeciton matrix.</returns>
        public Matrix getProjectionMatrix()
        {
            updateMatrix();
            return m_projMatrix;
        }

        /// <summary>
        /// Update the projection matrix after the size of the window has changed.
        /// </summary>
        public void updateMatrices()
        {
            m_updateMatrix = true;
        }
    }
}