/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace BoxEdit3D
{
    /// <summary>
    /// Base class for the gizmos that can be dragged by the user
    /// to manipulate the boxes within the scene.
    /// </summary>
    abstract class Gizmo
    {
        /// <summary>
        /// The box on which the gizmo is used on. Also determines the
        /// position, scale and rotation of the gizmo models.
        /// </summary>
        protected Box m_box;

        /// <summary>
        /// The GraphicsDeviceManager is needed to obtain the GraphicsDevice
        /// during rendering.
        /// </summary>
        protected GraphicsDeviceManager m_graphics;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="graphics">The GraphicsDeviceManager is needed during rendering.</param>
        public Gizmo(GraphicsDeviceManager graphics)
        {
            m_graphics = graphics;
        }

        /// <summary>
        /// Sets the box on which this gizmo operates. Can be set to null to disable
        /// this gizmo.
        /// </summary>
        /// <param name="box">The box on which this gizmo should work on.</param>
        public void setUsedBox(Box box)
        {
            m_box = box;
        }

        /// <summary>
        /// Renders the gizmo.
        /// </summary>
        /// <param name="viewProj">The View and Projection Matrix used to render the gizmo.</param>
        public abstract void render(Matrix viewProj);

        /// <summary>
        /// Start dragging on the gizmo. The given gizmoPart number indentifies which
        /// part of the gizmo is dragged.
        /// </summary>
        /// <param name="gizmoPart">The number of the gizmo part that is dragged.</param>
        public abstract void startDrag(int gizmoPart);

        /// <summary>
        /// Apply the the changed done by the gizmo and issue a command that
        /// capsulates the changes.
        /// </summary>
        /// <returns>The command that capsulates the changes to the box.</returns>
        public abstract Command applyDrag();

        /// <summary>
        /// Discard the changes done by the user and reset the values of the
        /// used box.
        /// </summary>
        public abstract void discardDrag();

        /// <summary>
        /// React to the mouse moves of the user when in drag mode.
        /// </summary>
        public abstract void handleMouseInput(Matrix viewProjectionMatrix);

        /// <summary>
        /// Returns the number of gizmo parts that can be dragged.
        /// </summary>
        /// <returns>The number of parts of the gizmo that can be dragged.</returns>
        public abstract int getNumberOfGizmoParts();

        /// <summary>
        /// Returns the model of the given gizmo part. This model can then be used to
        /// check if it was picked.
        /// </summary>
        /// <param name="index">The index of the gizmo part.</param>
        /// <returns>The model of the given gizmo part.</returns>
        public abstract Model getGizmoPart(int index);

        /// <summary>
        /// Returns if the gizmo is currently attached to a box and
        /// can therefore be seen.
        /// </summary>
        /// <returns>True if the gizmo is rendered.</returns>
        public virtual bool isVisible()
        {
            return (m_box != null);
        }

        /// <summary>
        /// Returns the transform matrix for the models of this gizmo.
        /// </summary>
        /// <returns>Returns the transform (model) matrix for the gizmo part models.</returns>
        public virtual Matrix getTransformMatrix()
        {
            if (m_box != null) return m_box.getTransformMatrix();
            else return Matrix.Identity;
        }

        /// <summary>
        /// Undo any changes to the box done by the gizmo,
        /// as it has lost the focus.
        /// </summary>
        public virtual void lostFocus()
        {
            setUsedBox(null);
        }
    }
}
