using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Graphics;

namespace BoxEdit3D
{
    /// <summary>
    /// Source code in this class is partly taken from the "XNA Triangle Accuracy Picking"
    /// Sample and is covered by the "Microsoft Permissive License". Please read the
    /// "readme.txt" file in the project folder for more information.
    /// </summary>
    class Picking
    {
        // used to unproject vetices to calculate the picking ray
        private GraphicsDeviceManager m_graphics;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="graphics">The GraphicsDeviceManager needed to construct the picking ray.</param>
        public Picking(GraphicsDeviceManager graphics)
        {
            m_graphics = graphics;
        }

        /// <summary>
        /// Calculate the picking ray that runs through the mouse cursor into the scene.
        /// </summary>
        /// <param name="projectionMatrix">The projection matrix used to render the scene.</param>
        /// <param name="viewMatrix">The view matrix used to render the scene.</param>
        /// <returns>The Ray running through the cursor position into the scene in world coordinates.</returns>
        public Ray CalculatePickingRay(Matrix projectionMatrix, Matrix viewMatrix)
        {
            // get mouse position
            Vector2 position;
            MouseState mouseState = Mouse.GetState();
            position.X = mouseState.X;
            position.Y = mouseState.Y;

            // calculate inverse view matrix
            Matrix invViewMatrix = Matrix.Invert(viewMatrix);

            // Compute the vector of the pick ray in screen space
            Vector3 v, vPickRayDir, vPickRayOrig;
            v.X = ((((2.0f * position.X) / m_graphics.GraphicsDevice.Viewport.Width) - 1) / projectionMatrix.M11);
            v.Y = (-(((2.0f * position.Y) / m_graphics.GraphicsDevice.Viewport.Height) - 1) / projectionMatrix.M22);
            v.Z = -1.0f;

            // Transform the screen space pick ray into 3D space
            vPickRayDir.X = v.X * invViewMatrix.M11 + v.Y * invViewMatrix.M21 + v.Z * invViewMatrix.M31;
            vPickRayDir.Y = v.X * invViewMatrix.M12 + v.Y * invViewMatrix.M22 + v.Z * invViewMatrix.M32;
            vPickRayDir.Z = v.X * invViewMatrix.M13 + v.Y * invViewMatrix.M23 + v.Z * invViewMatrix.M33;
            vPickRayOrig.X = invViewMatrix.M41;
            vPickRayOrig.Y = invViewMatrix.M42;
            vPickRayOrig.Z = invViewMatrix.M43;

            // return the picking ray
            return new Ray(vPickRayOrig, vPickRayDir);
        }

        /// <summary>
        /// Runs a per-triangle picking algorithm over all the models in the scene,
        /// storing which triangle is currently under the cursor.
        /// </summary>
        public bool PerformPicking(Matrix projectionMatrix, Matrix viewMatrix, Scene scene, Gizmo gizmo, out int pickedBox, out int pickedGizmoPart, out Vector3 planeIntersection)
        {
            bool foundIntersection = false;
            pickedBox = -1;
            pickedGizmoPart = -1;
            planeIntersection = Vector3.Zero;

            // Look up a collision ray based on the current cursor position. See the
            // Picking Sample documentation for a detailed explanation of this.
            Ray pickingRay = CalculatePickingRay(projectionMatrix, viewMatrix);

            // Keep track of the closest object we have seen so far, so we can
            // choose the closest one if there are several models under the cursor.
            float closestIntersection = float.MaxValue;

            // Loop over all boxes
            for (int i = 0; i < scene.numberOfBoxes(); i++)
            {

                // Perform the ray to model intersection test.
                float? intersection = RayIntersectsModel(pickingRay, Box.cubeModel, scene.getBox(i).getTransformMatrix());

                // Do we have a per-triangle intersection with this model?
                if (intersection != null)
                {
                    // If so, is it closer than any other model we might have
                    // previously intersected?
                    if (intersection < closestIntersection)
                    {
                        closestIntersection = intersection.Value;
                        pickedBox = i;
                        foundIntersection = true;
                    }
                }
            }

            // Loop over gizmo parts
            if (gizmo.isVisible())
            {
                for (int i = 0; i < gizmo.getNumberOfGizmoParts(); i++)
                {

                    // Perform the ray to model intersection test.
                    float? intersection = RayIntersectsModel(pickingRay, gizmo.getGizmoPart(i), gizmo.getTransformMatrix());

                    // Do we have a per-triangle intersection with this model?
                    if (intersection != null)
                    {
                        // If so, is it closer than any other model we might have
                        // previously intersected?
                        if (intersection < closestIntersection)
                        {
                            closestIntersection = intersection.Value;
                            pickedBox = -1;
                            pickedGizmoPart = i;
                            foundIntersection = true;
                        }
                    }
                }
            }

            // if no box intersection can be found,
            // try to find an intersection with the ground plane
            if (!foundIntersection)
            {
                Plane plane = new Plane(new Vector4(0.0f, 1.0f, 0.0f, 0.0f));
                
                float distance;
                if (PlaneIntersectRay(ref plane, ref pickingRay, out distance))
                {
                    // did we find an intersection in front of the camera?
                    if (distance > 0)
                    {
                        planeIntersection = pickingRay.Position + (distance * pickingRay.Direction);
                        foundIntersection = true;
                    }
                }
            }

            return foundIntersection;
        }

        /// <summary>
        /// Checks if the given ray intersects with the given plane. Returns
        /// the distance on the ray to the plane.
        /// </summary>
        /// <param name="p">Plane to be intersect the ray with.</param>
        /// <param name="r">Ray that is used to intersect with the plane.</param>
        /// <param name="fDist">The distance on the ray to the plane.</param>
        /// <returns></returns>
        public static bool PlaneIntersectRay(ref Plane p, ref Ray r, out float fDist)
        {
            // set distance to 0
            fDist = 0.0f;

            // check for parallel ray
            float denom = Vector3.Dot(p.Normal, r.Direction);
            if (denom == 0.0f) return false;

            // calculate distance
            float numer = Vector3.Dot(p.Normal, r.Position);
            fDist = -(((numer + p.D)) / denom);
            return true;
        }

        /// <summary>
        /// Checks whether a ray intersects a model. This method needs to access
        /// the model vertex data, so the model must have been built using the
        /// custom TrianglePickingProcessor provided as part of this solution.
        /// Returns the distance along the ray to the point of intersection, or null
        /// if there is no intersection.
        /// </summary>
        public static float? RayIntersectsModel(Ray ray, Model model, Matrix transformMatrix)
        {
            // The input ray is in world space, but our model data is stored in object
            // space. We would normally have to transform all the model data by the
            // modelTransform matrix, moving it into world space before we test it
            // against the ray. That transform can be slow if there are a lot of
            // triangles in the model, however, so instead we do the opposite.
            // Transforming our ray by the inverse modelTransform moves it into object
            // space, where we can test it directly against our model data. Since there
            // is only one ray but typically many triangles, doing things this way
            // around can be much faster.

            Matrix inverseTransform = Matrix.Invert(transformMatrix);

            ray.Position = Vector3.Transform(ray.Position, inverseTransform);
            ray.Direction = Vector3.TransformNormal(ray.Direction, inverseTransform);

            // Look up our custom collision data from the Tag property of the model.
            Dictionary<string, object> tagData = (Dictionary<string, object>)model.Tag;

            if (tagData == null)
            {
                throw new InvalidOperationException(
                    "Model.Tag is not set correctly. Make sure your model " +
                    "was built using the custom TrianglePickingProcessor.");
            }

            // we cannot use a bounding sphere test, as it does not work with
            // scaling the model

            // Keep track of the closest triangle we found so far,
            // so we can always return the closest one.
            float? closestIntersection = null;

            // Loop over the vertex data, 3 at a time (3 vertices = 1 triangle).
            Vector3[] vertices = (Vector3[])tagData["Vertices"];

            for (int i = 0; i < vertices.Length; i += 3)
            {
                // Perform a ray to triangle intersection test.
                float? intersection;

                RayIntersectsTriangle(ref ray,
                                      ref vertices[i],
                                      ref vertices[i + 1],
                                      ref vertices[i + 2],
                                      out intersection);

                // Does the ray intersect this triangle?
                if (intersection != null)
                {
                    // If so, is it closer than any other previous triangle?
                    if ((closestIntersection == null) ||
                        (intersection < closestIntersection))
                    {
                        // Store the distance to this triangle.
                        closestIntersection = intersection;
                    }
                }
            }

            return closestIntersection;
        }


        /// <summary>
        /// Checks whether a ray intersects a triangle. This uses the algorithm
        /// developed by Tomas Moller and Ben Trumbore, which was published in the
        /// Journal of Graphics Tools, volume 2, "Fast, Minimum Storage Ray-Triangle
        /// Intersection".
        /// 
        /// This method is implemented using the pass-by-reference versions of the
        /// XNA math functions. Using these overloads is generally not recommended,
        /// because they make the code less readable than the normal pass-by-value
        /// versions. This method can be called very frequently in a tight inner loop,
        /// however, so in this particular case the performance benefits from passing
        /// everything by reference outweigh the loss of readability.
        /// </summary>
        public static void RayIntersectsTriangle(ref Ray ray,
                                          ref Vector3 vertex1,
                                          ref Vector3 vertex2,
                                          ref Vector3 vertex3, out float? result)
        {
            // Compute vectors along two edges of the triangle.
            Vector3 edge1, edge2;

            Vector3.Subtract(ref vertex2, ref vertex1, out edge1);
            Vector3.Subtract(ref vertex3, ref vertex1, out edge2);

            // Compute the determinant.
            Vector3 directionCrossEdge2;
            Vector3.Cross(ref ray.Direction, ref edge2, out directionCrossEdge2);

            float determinant;
            Vector3.Dot(ref edge1, ref directionCrossEdge2, out determinant);

            // If the ray is parallel to the triangle plane, there is no collision.
            if (determinant > -float.Epsilon && determinant < float.Epsilon)
            {
                result = null;
                return;
            }

            float inverseDeterminant = 1.0f / determinant;

            // Calculate the U parameter of the intersection point.
            Vector3 distanceVector;
            Vector3.Subtract(ref ray.Position, ref vertex1, out distanceVector);

            float triangleU;
            Vector3.Dot(ref distanceVector, ref directionCrossEdge2, out triangleU);
            triangleU *= inverseDeterminant;

            // Make sure it is inside the triangle.
            if (triangleU < 0 || triangleU > 1)
            {
                result = null;
                return;
            }

            // Calculate the V parameter of the intersection point.
            Vector3 distanceCrossEdge1;
            Vector3.Cross(ref distanceVector, ref edge1, out distanceCrossEdge1);

            float triangleV;
            Vector3.Dot(ref ray.Direction, ref distanceCrossEdge1, out triangleV);
            triangleV *= inverseDeterminant;

            // Make sure it is inside the triangle.
            if (triangleV < 0 || triangleU + triangleV > 1)
            {
                result = null;
                return;
            }

            // Compute the distance along the ray to the triangle.
            float rayDistance;
            Vector3.Dot(ref edge2, ref distanceCrossEdge1, out rayDistance);
            rayDistance *= inverseDeterminant;

            // Is the triangle behind the ray origin?
            if (rayDistance < 0)
            {
                result = null;
                return;
            }

            result = rayDistance;
        }
    }
}
