/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Input;
using System.Windows.Forms;

namespace BoxEdit3D
{
    /// <summary>
    /// This class is used for most of the User Interface logic of the
    /// BoxEdit3D project. It stores the current gizmo mode and reacts
    /// to the user input.
    /// </summary>
    class UserInput
    {
        // callback is used to tell the main application to shut down
        public delegate void simpleCallback();

        // instances of the different classes that are used
        // to store information about the curent state of the editor
        private Camera m_camera;
        private Scene m_scene;
        private CommandHistory m_commandHistory;
        private Picking m_picking;
        private GraphicsDeviceManager m_graphics;

        // variables to detect a double mouse clicks
        private bool m_leftMouseButtonPressed = false;
        private bool m_detectedDoubleClick = false;
        private DateTime m_lastMouseClick = DateTime.MinValue;
        private int m_lastMouseClickPosX = -1;
        private int m_lastMouseClickPosY = -1;

        // variables used to sperate each individual key press
        private bool m_undoButtonDown = false;
        private bool m_redoButtonDown = false;
        private bool m_delButtonDown = false;
        private bool m_resetViewButtonDown = false;
        private bool m_resetSceneButtonDown = false;

        // gizmo instances
        private Gizmo m_currentGizmo;
        private DummyGizmo m_dummyGizmo;
        private TranslationGizmo m_translationGizmo;
        private RotationGizmo m_rotationGizmo;
        private ScaleGizmo m_scaleGizmo;

        // variables used for the open- and save filedialogs
        private System.Windows.Forms.OpenFileDialog m_xmlOpenFileDialog;
        private System.Windows.Forms.SaveFileDialog m_xmlSaveFileDialog;
        private System.Windows.Forms.OpenFileDialog m_textureOpenFileDialog;
        private bool m_xmlLoadButtonDown = false;
        private bool m_xmlSaveButtonDown = false;
        private bool m_textureLoadButtonDown = false;

        // callback method to tell the application to shut down
        private simpleCallback m_exitCallback;

        // enumeration of the different gizmo modes
        private enum GizmoMode
        {
            NoGizmo = 0,
            TranslationGizmo,
            RotationGizmo,
            ScaleGizmo
        }

        // current gizmo mode what kind of gizmo is displayed
        // on the currently selected box
        private GizmoMode m_currentMode = GizmoMode.NoGizmo;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="graphics">The GraphicsDeviceManager needed to initialize certain classes.</param>
        /// <param name="scene">The instance of the scene class that the editor works on.</param>
        /// <param name="exitCallback">This callback method is called when the user wants to exit the editor.</param>
        public UserInput(GraphicsDeviceManager graphics, Scene scene, simpleCallback exitCallback)
        {
            // initialize
            m_graphics = graphics;
            m_camera = new Camera(m_graphics);
            m_scene = scene;
            m_picking = new Picking(m_graphics);
            m_commandHistory = new CommandHistory();
            m_translationGizmo = new TranslationGizmo(m_graphics);
            m_rotationGizmo = new RotationGizmo(m_graphics);
            m_scaleGizmo = new ScaleGizmo(m_graphics);
            m_dummyGizmo = new DummyGizmo(m_graphics);
            m_currentGizmo = m_dummyGizmo;
            m_currentMode = GizmoMode.NoGizmo;
            m_exitCallback = exitCallback;

            // initialize file dialogs
            m_xmlOpenFileDialog = new System.Windows.Forms.OpenFileDialog(); ;
            m_xmlSaveFileDialog = new System.Windows.Forms.SaveFileDialog();
            m_textureOpenFileDialog = new System.Windows.Forms.OpenFileDialog(); ;

            // set possible extensions
            m_xmlOpenFileDialog.DefaultExt = "xml";
            m_xmlOpenFileDialog.Filter = "XML files(*.xml)|*.xml";
            m_xmlSaveFileDialog.DefaultExt = "xml";
            m_xmlSaveFileDialog.Filter = "XML files(*.xml)|*.xml";
            m_textureOpenFileDialog.Filter = "All supported file formats|*.bmp;*.dds;*.dib;*.hdr;*.jpg;*pfm;*.png;*.ppm;*.tga|"
                                           + "Windows bitmap files (*.bmp, *.dib)|*.bmp;*.dib|"
                                           + "DirectDrawSurface files (*.dds)|*.dds|"
                                           + "High dynamic-range files (*.hdr)|*.hdr|"
                                           + "JPEG files (*.jpg)|*.jpg|"
                                           + "Portable float map files (*.pfm)|*.pfm|"
                                           + "Portable Network Graphics files (*.png)|*.png|"
                                           + "Portable pixmap files (*.ppm)|*.ppm|"
                                           + "Truevision Targa image files (*.tga)|*.tga";
        }

        /// <summary>
        /// Main method of this class. This method should be called every frame
        /// so that the user can interact with the editor. Handles mouse and
        /// keyboard input.
        /// </summary>
        /// <param name="time">The game time object that stores the passed time since the last frame.</param>
        public void Update(GameTime time)
        {
            // let the camera update itself based on the mouse input
            m_camera.handleMouseInput();

            // update camera position
            KeyboardState keyboardState = Keyboard.GetState();
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.W))
            {
                m_camera.moveForward();
            }
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.S))
            {
                m_camera.moveBackward();
            }
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.A))
            {
                m_camera.strafeLeft();
            }
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.D))
            {
                m_camera.strafeRight();
            }
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.Space))
            {
                m_camera.moveUp();
            }
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.LeftShift))
            {
                m_camera.moveDown();
            }

            // check for escape to exit application
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.Escape))
            {
                m_exitCallback();
            }

            // check for undo + redo
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.LeftControl))
            {
                if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.Z))
                {
                    if (!m_undoButtonDown)
                    {
                        // undo one command
                        m_commandHistory.undoCommand();
                        updateGizmo();
                        m_undoButtonDown = true;
                    }
                }
            }
            if (keyboardState.IsKeyUp(Microsoft.Xna.Framework.Input.Keys.Z)) m_undoButtonDown = false;
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.LeftControl))
            {
                if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.Y))
                {
                    if (!m_redoButtonDown)
                    {
                        // redo one command
                        m_commandHistory.redoCommand();
                        updateGizmo();
                        m_redoButtonDown = true;
                    }
                }
            }
            if (keyboardState.IsKeyUp(Microsoft.Xna.Framework.Input.Keys.Y)) m_redoButtonDown = false;

            // check for gizmo mode change
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.D1)) updateGizmoMode(GizmoMode.NoGizmo);
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.D2)) updateGizmoMode(GizmoMode.TranslationGizmo);
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.D3)) updateGizmoMode(GizmoMode.RotationGizmo);
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.D4)) updateGizmoMode(GizmoMode.ScaleGizmo);

            // check if file dialogs should be displayed
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.L))
            {
                if (!m_xmlLoadButtonDown)
                {
                    m_xmlLoadButtonDown = true;
                    // open xml load file dialog
                    // toggle out of full screen to show dialog
                    bool isFullScreen = m_graphics.IsFullScreen;
                    if (isFullScreen) m_graphics.ToggleFullScreen();
                    if (m_graphics.IsFullScreen) m_graphics.ToggleFullScreen();
                    if (m_xmlOpenFileDialog.ShowDialog() == DialogResult.OK)
                    {
                        if (m_scene.loadSceneFromXML(m_graphics, m_xmlOpenFileDialog.FileName))
                        {
                            // reset CommandHistory
                            m_commandHistory.ClearHistory();
                        }
                        
                        if (isFullScreen) m_graphics.ToggleFullScreen();
                        return;
                    }
                    if (isFullScreen) m_graphics.ToggleFullScreen();
                }
            }
            if (keyboardState.IsKeyUp(Microsoft.Xna.Framework.Input.Keys.L)) m_xmlLoadButtonDown = false;
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.K))
            {
                if (!m_xmlSaveButtonDown)
                {
                    m_xmlSaveButtonDown = true;
                    // open xml save file dialog
                    // toggle out of full screen to show dialog
                    bool isFullScreen = m_graphics.IsFullScreen;
                    if (isFullScreen) m_graphics.ToggleFullScreen();
                    if (m_xmlSaveFileDialog.ShowDialog() == DialogResult.OK)
                    {
                        m_scene.saveSceneToXML(m_xmlSaveFileDialog.FileName);
                        if (isFullScreen) m_graphics.ToggleFullScreen();
                        return;
                    }
                    if (isFullScreen) m_graphics.ToggleFullScreen();
                }
            }
            if (keyboardState.IsKeyUp(Microsoft.Xna.Framework.Input.Keys.K)) m_xmlSaveButtonDown = false;
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.T))
            {
                if (!m_textureLoadButtonDown)
                {
                    m_textureLoadButtonDown = true;
                    // open texture load file dialog
                    // check if a box is selected
                    if (m_scene.isBoxSelected())
                    {
                        bool isFullScreen = m_graphics.IsFullScreen;
                        if (isFullScreen) m_graphics.ToggleFullScreen();
                        if (m_textureOpenFileDialog.ShowDialog() == DialogResult.OK)
                        {
                            TextureCommand command = new TextureCommand(m_scene.getSelectedBox(), m_textureOpenFileDialog.FileName);
                            m_commandHistory.executeAndAddCommand(command);
                        }
                        if (isFullScreen) m_graphics.ToggleFullScreen();
                    }
                }
            }
            if (keyboardState.IsKeyUp(Microsoft.Xna.Framework.Input.Keys.T)) m_textureLoadButtonDown = false;

            // check if user wants to delete the selected box
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.Delete))
            {
                if (!m_delButtonDown)
                {
                    m_delButtonDown = true;
                    // delete selected box
                    if (m_scene.isBoxSelected())
                    {
                        DestroyCommand command = new DestroyCommand(m_scene.getSelectedBox(), m_scene);
                        m_commandHistory.executeAndAddCommand(command);

                        // update gizmos
                        updateGizmo();
                    }
                }
            }
            if (keyboardState.IsKeyUp(Microsoft.Xna.Framework.Input.Keys.Delete)) m_delButtonDown = false;

            // check if the user wants to reset the scene
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.N))
            {
                if (!m_resetSceneButtonDown)
                {
                    m_resetSceneButtonDown = true;
                    // reset scene
                    m_scene.reset();
                    m_commandHistory.ClearHistory();
                    return;
                }
            }
            if (keyboardState.IsKeyUp(Microsoft.Xna.Framework.Input.Keys.N)) m_resetSceneButtonDown = false;

            // check if the user wants to reset the view
            if (keyboardState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.R))
            {
                if (!m_resetViewButtonDown)
                {
                    m_resetViewButtonDown = true;
                    // reset scene
                    m_camera.reset();
                    return;
                }
            }
            if (keyboardState.IsKeyUp(Microsoft.Xna.Framework.Input.Keys.R)) m_resetViewButtonDown = false;

            // check if picking is needed
            MouseState mouseState = Mouse.GetState();
            if ((m_leftMouseButtonPressed) && (mouseState.LeftButton == Microsoft.Xna.Framework.Input.ButtonState.Pressed))
            {
                // User holds mousebutton down
                // if user is dragging update gizmo
                m_currentGizmo.handleMouseInput(m_camera.getViewProjectionMatrix());
            }
            else if ((!m_leftMouseButtonPressed) && (mouseState.LeftButton == Microsoft.Xna.Framework.Input.ButtonState.Pressed))
            {
                // User presses mousebutton down
                m_leftMouseButtonPressed = true;

                // user wants to Pick something, starts an double click or finishes an double click
                // check for double click
                TimeSpan duration = DateTime.Now - m_lastMouseClick;

                // find out picked box / picked gizmo part / picked location
                int boxNumber;
                int gizmoPartNumber;
                Vector3 planeIntersection;
                bool pickedBox = false;
                bool pickedGizmo = false;
                bool pickedLocation = false;
                if (m_picking.PerformPicking(m_camera.getProjectionMatrix(), m_camera.getViewMatrix(), m_scene, m_currentGizmo, out boxNumber, out gizmoPartNumber, out planeIntersection)) 
                {
                    if (boxNumber != -1) pickedBox = true;
                    else if (gizmoPartNumber != -1) pickedGizmo = true;
                    else pickedLocation = true;                   
                };
                
                // check if mouse has moved since last click
                // max half second for an double click
                if ((mouseState.X == m_lastMouseClickPosX) && (mouseState.Y == m_lastMouseClickPosY) && (duration.TotalMilliseconds <= 350.0f)) 
                {
                    // found double click
                    // prevent second double click within 350 ms
                    m_detectedDoubleClick = true;

                    // only create box if a location was picked
                    if (pickedLocation)
                    {
                        // create new box
                        Box box = new Box(m_graphics);
                        box.setTranslation(planeIntersection);

                        // issue a create command and add it to the CommandHistory
                        CreateCommand command = new CreateCommand(box, m_scene);
                        m_commandHistory.executeAndAddCommand(command);

                        // update gizmos
                        updateGizmo();
                    }
                    else if (pickedBox)
                    {
                        // just select the picked box
                        selectBox(boxNumber);
                    }                  
                }
                else
                {
                    // not a double click -> select box if picked
                    if (pickedBox) selectBox(boxNumber);
                    else if (pickedGizmo) m_currentGizmo.startDrag(gizmoPartNumber);
                    else
                    {
                        // deselect box
                        selectBox(-1);
                        updateGizmo();
                    }
                }
            }
            else if ((m_leftMouseButtonPressed) && (mouseState.LeftButton == Microsoft.Xna.Framework.Input.ButtonState.Released))
            {
                m_lastMouseClickPosX = mouseState.X;
                m_lastMouseClickPosY = mouseState.Y;

                // User has released mousebutton
                if (m_detectedDoubleClick)
                {
                    // prevent next click to be a double click
                    m_lastMouseClick = DateTime.MinValue;
                    m_detectedDoubleClick = false;
                }
                else m_lastMouseClick = DateTime.Now;

                m_leftMouseButtonPressed = false;

                // gizmo changes should be applied, command should be issued to the CommandHistory
                Command command = m_currentGizmo.applyDrag();
                if (command != null)
                {
                    m_commandHistory.executeAndAddCommand(command);
                }
            }
        }

        /// <summary>
        /// This method selects the indicated box in the scene. Also handles
        /// the correct usage of the gizmos.
        /// </summary>
        /// <param name="boxNumber">The box number that should be selected.</param>
        private void selectBox(int boxNumber)
        {
            // if already a box was selected,
            // undo any changes by the gizmo
            gizmoLostFocus();

            // select new box
            m_scene.selectBox(boxNumber);

            // update gizmo
            gizmoGotNewBox();
        }

        /// <summary>
        /// Make sure that the current gizmo is up-to-date.
        /// </summary>
        private void updateGizmo()
        {
            gizmoLostFocus();
            gizmoGotNewBox();
        }

        /// <summary>
        /// The current gizmo lost the focus and must be notified.
        /// </summary>
        private void gizmoLostFocus()
        {
            m_currentGizmo.lostFocus();
        }

        /// <summary>
        /// A new box has been selected and the gizmo must therefor be updated.
        /// </summary>
        private void gizmoGotNewBox()
        {
            // check if a box is selected
            if (m_scene.isBoxSelected())
            {
                Box box = m_scene.getSelectedBox();
                // initialize new gizmo with selected box
                m_currentGizmo.setUsedBox(box);
            }
            else gizmoLostFocus();
        }

        /// <summary>
        /// Set the current gizmo mode to the indicated new gizmo mode.
        /// </summary>
        /// <param name="mode">The gizmo mode that should be set.</param>
        private void updateGizmoMode(GizmoMode mode)
        {
            if (m_currentMode != mode) {
                // clean up last gizmo
                gizmoLostFocus();

                // set new mode
                m_currentMode = mode;

                switch (m_currentMode)
                {
                    case GizmoMode.NoGizmo:
                        m_currentGizmo = m_dummyGizmo;
                        break;
                    case GizmoMode.TranslationGizmo:
                        m_currentGizmo = m_translationGizmo;
                        break;
                    case GizmoMode.RotationGizmo:
                        m_currentGizmo = m_rotationGizmo;
                        break;
                    case GizmoMode.ScaleGizmo:
                        m_currentGizmo = m_scaleGizmo;
                        break;
                }

                // check if a gizmo gets a new box
                gizmoGotNewBox();
            }
        }

        /// <summary>
        /// Returns the view matrix used to render the scene.
        /// </summary>
        /// <returns>The view matrix that is used to render the scene.</returns>
        public Matrix getViewMatrix()
        {
            return m_camera.getViewMatrix();
        }
        
        /// <summary>
        /// Returns the projection matrix used to render the scene.
        /// </summary>
        /// <returns>The projection matrix that is used to render the scene.</returns>
        public Matrix getProjectionMatrix()
        {
            return m_camera.getProjectionMatrix();
        }

        /// <summary>
        /// Returns the combined view and projection matrix used to render the scene.
        /// </summary>
        /// <returns>The view and projection matrix that is used to render the scene.</returns>
        public Matrix getViewProjectionMatrix()
        {
            return m_camera.getViewProjectionMatrix();
        }

        /// <summary>
        /// Render the current gizmo.
        /// </summary>
        /// <param name="viewProjMatrix">The view and projection matrix used to render the gizmo.</param>
        public void renderGizmo(Matrix viewProjMatrix)
        {
            if (m_scene.isBoxSelected())
            {
                m_currentGizmo.render(viewProjMatrix);
            }
        }

        /// <summary>
        /// Refresh camera matrices after a resize of the window.
        /// </summary>
        public void updateCameraMatrices()
        {
            m_camera.updateMatrices();
        }
    }
}
